import datetime
from collections import OrderedDict
from collections import defaultdict
from functools import partial

from gisdo import constants as const
from gisdo import xml_helpers
from gisdo.exceptions import NoDataInMo
from gisdo.index import get_enrolled_age_filter
from gisdo.index.direct import ind_19_3
from gisdo.index.direct import DirectIndex
from gisdo.algorithm.direct import get_children_last_direct_ids
from gisdo.algorithm.direct import get_32_filter
from gisdo.algorithm.direct import get_321_filter
from gisdo.algorithm.enrolled import get_28_index

from gisdo.index.direct import ind_29_1
from gisdo.index.direct import ind_29_2
from gisdo.index.direct import ind_29_3
from gisdo.index.capacities import childrens_in_age_category
from gisdo.index.enrolled import Index
from gisdo.index.enrolled import get_age_for_date_filter
from gisdo.models import ScheduleSettings
from gisdo.service import FilialGroupData
from gisdo.service import TemplateView
from gisdo.service import GroupData
from gisdo.service import ParentGroupData
from gisdo.xml_helpers import get_detailed_passport
from gisdo.utils import DateMixin
from kinder.core.children.models import Children
from kinder.core.dict.models import GroupTypeEnumerate
from kinder.core.dict.models import UnitKind
from kinder.core.group.models import Group
from kinder.core.group.models import GroupStatusEnum
from kinder.core.helpers import get_instance
from kinder.core.unit.models import Unit
from .helpers import get_count_data

from ..collector import DataCollector
from .constants import *
from .helpers import prep_ind_19_3_queries
from .helpers import prep_ind_29_1_queries
from .helpers import prepare_direct_index_queries
from .helpers import prep_ind_29_2_query
from .helpers import prep_in_29_2_data
from .helpers import prep_ind_29_3_query
from .helpers import prep_directs_in_group_queries
from .helpers import prepare_enrolled_index_queries
from .helpers import prep_ind_enrolled_data
from .helpers import ReportQuery
from ..index import get_queue_index_collection
from ..index.queue import get_age_filter


class ReportDataCollector(DataCollector):
    """Сборщик данных отчета "Выгрузка детей по показателю/тегу"."""

    # Название в словаре self._cache для данных о выгружаемых в xml тегах
    XML_TAGS = 'xml_tags'

    def __init__(self, mo, unit_helper, report_start_date, indexes=None):
        super().__init__(mo, unit_helper, report_start_date)

        self.indexes = indexes or []
        self._cache[self.XML_TAGS] = {}

    def collect(self, report_form=None):

        # Признак наличия садов по которым можно собрать данные
        dou_list = self._unit_helper.get_mo_units(self._mo.id)
        have_data = len(dou_list)

        if not have_data:
            raise NoDataInMo(
                f'Нет садов в МО {self._mo} пригодных для отправки'
            )

        # Для расчета показателя 32 и 32.1, 32.2, необходим список
        # идентификаторов последних направлений у ребенка в рамках МО
        if INDEX_32 in self.indexes:
            self.list_last_id_directs_32 = get_children_last_direct_ids(
                self._mo.id, self._report_start_date, get_32_filter())
        if INDEX_32_1 in self.indexes:
            self.list_last_id_directs_321 = get_children_last_direct_ids(
                self._mo.id, self._report_start_date, get_321_filter())

        group_data = dict()

        # Собираем по ДОУ
        for dou in dou_list:
            self._collect_enrolled_data(dou)
            self._collect_capacities(dou)

            filials = xml_helpers.get_dou_filials(
                dou.get_mo()
            ).select_related("filial")

            parent_group_data = ParentGroupData(
                dou.get_mo(), self._report_start_date, filials,
                prep_enrollments_func=prep_ind_enrolled_data
            )

            filial_group_data = FilialGroupData(
                dou, parent_group_data, self._report_start_date,
                prep_enrollments_func=prep_ind_enrolled_data
            )

            group_data[dou.id] = list(filial_group_data.groups)

        self._cache[self.CAPACITIES] = group_data
        if self.has_selected_index(XML_TAGS):
            self._cache[self.XML_TAGS] = self._calculate_xml_tags(report_form)
        return self._cache

    @staticmethod
    def _merge_dicts(first_dict, second_dict):
        if not first_dict or not second_dict:
            return

        first_dou_id = list(first_dict.keys())[0]
        second_dou_id = list(second_dict.keys())[0]

        for index, first_index_dict in first_dict[first_dou_id].items():
            second_index_dict = second_dict[second_dou_id][index]
            age_categories = set(first_index_dict.keys()) | set(
                second_index_dict.keys())

            for age_cat in age_categories:
                if age_cat not in second_index_dict:
                    continue

                data = second_index_dict[age_cat]
                if age_cat in first_index_dict:
                    first_index_dict[age_cat].extend(data)
                else:
                    first_index_dict[age_cat] = data.copy()

    def get_enrolled_data(self, collection, age_filter):
        """Фильтрация данных по тегу"""

        data = [data for data in collection if age_filter(*data)]

        return data

    def get_enrolled_index_data(self, _unit, report_start_date):
        """Получение данных для тега enrolled"""

        full_enrolled_data = defaultdict(list)

        filials = xml_helpers.get_dou_filials(
            _unit
        ).select_related("filial")

        parent_group_data = ParentGroupData(
            _unit, report_start_date, filials,
            prep_enrollments_func=prep_ind_enrolled_data
        )

        # головная организация передается только если у нее есть группы
        if xml_helpers.has_fact_groups(_unit):
            for parent_group in Group.objects.filter(
                    status=GroupStatusEnum.FACT, unit=_unit).order_by('id'):
                age_from, age_to = GroupData.get_age_range(parent_group)
                age_cat = f'от {int(age_from)} до {int(age_to)} лет'

                enrolled = parent_group_data.get_enrollments_count(
                    parent_group,
                    prep_enrollments_func=prep_ind_enrolled_data)

                full_enrolled_data[age_cat].extend(enrolled)

        for filial in (fd.filial for fd in filials.iterator()):
            # филиал передается только если у него есть группы
            if xml_helpers.has_fact_groups(filial):

                filial_group_data = FilialGroupData(
                    filial, parent_group_data, report_start_date,
                    prep_enrollments_func=prep_ind_enrolled_data
                )

                for filial_group in Group.objects.filter(
                        status=GroupStatusEnum.FACT, unit=_unit).order_by('id'):
                    age_from, age_to = GroupData.get_age_range(filial_group)
                    age_cat = f'от {int(age_from)} до {int(age_to)} лет'

                    enrolled = filial_group_data.get_enrollments_count(
                        filial_group,
                        prep_enrollments_func=prep_ind_enrolled_data)

                    full_enrolled_data[age_cat].extend(enrolled)

        return full_enrolled_data

    def has_selected_index(self, indexes):
        """Проверяет, был ли выбран хотя бы один из указанных индексов

        :param indexes: Названия индексов для проверки
        :type indexes: Iterable

        :return: Был ли выбран хотя бы один из указанных индексов для отчёта
        :rtype: bool
        """
        return any(index in self.indexes for index in indexes)

    def _collect_enrolled_data(self, dou):

        def _inner(_dou, report_start_date, calc_filial=False):
            enrolled = {
                _dou.id: OrderedDict([
                    (INDEX_ENROLLED, {}),
                    (INDEX_12, {}),
                    (INDEX_8, {}),
                    (INDEX_8_1, {}),
                    (INDEX_8_2, {}),
                    (INDEX_8_3, {}),
                    (INDEX_18_1, {}),
                    (INDEX_19, {}),
                    (INDEX_19_1, {}),
                    (INDEX_19_2, {}),
                    (INDEX_19_3, {}),
                    (INDEX_20, {}),
                    (INDEX_20_1, {}),
                    (INDEX_20_2, {}),
                    (INDEX_20_5, {}),
                    (INDEX_20_6, {}),
                    (INDEX_20_7, {}),
                    (INDEX_20_8, {}),
                    (INDEX_24, {}),
                    (INDEX_28, {}),
                    (INDEX_29, {}),
                    (INDEX_29_1, {}),
                    (INDEX_29_2, {}),
                    (INDEX_29_3, {}),
                    (INDEX_31, {}),
                    (INDEX_32, {}),
                    (INDEX_32_1, {}),
                ])
            }
            dou_enrolled = enrolled[_dou.id]

            if INDEX_ENROLLED in self.indexes:
                ind_enrolled_data = self.get_enrolled_index_data(
                    _dou, report_start_date)
                if ind_enrolled_data:
                    dou_enrolled[INDEX_ENROLLED] = ind_enrolled_data

            enrolled_index = Index(_dou, report_start_date)
            prepared_enrolled_index = partial(
                enrolled_index,
                prep_enrolled_func=prepare_enrolled_index_queries,
            )

            if INDEX_29_1 in self.indexes:
                index_29_1 = ind_29_1(_dou, self._report_start_date,
                                      prep_func=prep_ind_29_1_queries)

            queue_indexes = {}

            for index in QUEUE_INDEXES:
                if index in self.indexes:
                    params = QUEUE_INDEXES_PARAMS.get(index, {})
                    queue_indexes[index] = get_queue_index_collection(
                        index, _dou, report=True, **params)
            if INDEX_29_2 in self.indexes:
                index_29_2 = ind_29_2(_dou, prep_func=prep_ind_29_2_query)

            if INDEX_29_3 in self.indexes:
                index_29_3 = ind_29_3(_dou, prep_func=prep_ind_29_3_query)

            if not calc_filial:
                if self.has_selected_index([INDEX_19, INDEX_19_1, INDEX_19_2]):
                    index_19 = list(prepared_enrolled_index())

                if INDEX_24 in self.indexes:
                    index_24 = list(prepared_enrolled_index(
                        group_filter=lambda group: (
                            group.type and
                            group.type.code == GroupTypeEnumerate.HEALTH
                        )
                    ))

                if INDEX_28 in self.indexes:
                    index_28 = get_28_index(prepared_enrolled_index)

            for age_cat in list(const.AGE_CATEGORIES_FULL.values()):
                current_year_enrolled_filter = get_enrolled_age_filter(
                    age_cat, DateMixin.get_current_calendar_year())
                on_date_enrolled_filter = get_enrolled_age_filter(
                    age_cat, datetime.date.today)
                next_calendar_year_enrolled = get_enrolled_age_filter(
                    age_cat, DateMixin.get_next_calendar_year)
                on_other_date_age_filter = get_age_for_date_filter(age_cat)

                on_date_filter = get_age_filter(age_cat, datetime.date.today)
                # Общая численность детей зачисленных в ДОО
                if not calc_filial:
                    if INDEX_19 in self.indexes:
                        ind_19_data = self.get_enrolled_data(
                            index_19, current_year_enrolled_filter)
                        if ind_19_data:
                            dou_enrolled[INDEX_19][age_cat] = ind_19_data

                    if INDEX_19_1 in self.indexes:
                        ind_19_1_data = self.get_enrolled_data(
                            index_19, on_date_enrolled_filter)
                        if ind_19_1_data:
                            dou_enrolled[INDEX_19_1][age_cat] = ind_19_1_data

                    if INDEX_19_2 in self.indexes:
                        ind_19_2_data = self.get_enrolled_data(
                            index_19, next_calendar_year_enrolled)
                        if ind_19_2_data:
                            dou_enrolled[INDEX_19_2][age_cat] = ind_19_2_data

                if INDEX_19_3 in self.indexes:
                    ind_19_3_children_ids = ind_19_3(
                        _dou, report_start_date, age_range=age_cat,
                        prep_direct_func=prep_ind_19_3_queries)
                    ind_19_3_data = ReportQuery.get_child_query(
                        Children.objects.filter(id__in=ind_19_3_children_ids))
                    if ind_19_3_data:
                        dou_enrolled[INDEX_19_3][age_cat] = list(ind_19_3_data)

                # Общее количество детей зачисленных в
                # оздоровительные и компенсирующие группы
                if not calc_filial:
                    if INDEX_24 in self.indexes:
                        dou_enrolled[INDEX_24][age_cat] = self.get_enrolled_data(
                            index_24, on_date_enrolled_filter)

                    # Общая численность детей, посещающих ДОО в режиме
                    # кратковременного пребывания
                    if INDEX_28 in self.indexes:
                        dou_enrolled[INDEX_28][age_cat] = self.get_enrolled_data(
                            index_28, on_date_enrolled_filter)

                if INDEX_29_1 in self.indexes:
                    ind_29_1_data = self.get_enrolled_data(
                        index_29_1, on_date_enrolled_filter)
                    if ind_29_1_data:
                        dou_enrolled[INDEX_29_1][age_cat] = ind_29_1_data

                if INDEX_29_2 in self.indexes:
                    index_29_2_data = self.get_enrolled_data(
                        index_29_2, on_other_date_age_filter)
                    if index_29_2_data:
                        dou_enrolled[INDEX_29_2][age_cat] = prep_in_29_2_data(
                            index_29_2_data)

                if INDEX_29_3 in self.indexes:
                    index_29_3_data = self.get_enrolled_data(
                        index_29_3, next_calendar_year_enrolled)
                    if index_29_3_data:
                        dou_enrolled[INDEX_29_3][age_cat] = index_29_3_data

                for index in QUEUE_INDEXES:
                    if index in self.indexes:
                        dou_enrolled[index][age_cat] = get_count_data(
                            queue_indexes[index], on_date_filter)

            # Общее количество детей зачисленных в семейные,
            # оздоровительные и компенсирующие группы
            if not calc_filial and INDEX_29 in self.indexes:
                ind_29_data = list(prepared_enrolled_index(
                    query_filter='predictable_decrease_contingent'))

                if ind_29_data:
                    dou_enrolled[INDEX_29][const.ALL] = ind_29_data

            return enrolled

        enrolled_data = _inner(dou, self._report_start_date)

        for filial in self._unit_helper.get_filial(dou).iterator():
            filial_enrolled = _inner(
                filial, self._report_start_date, calc_filial=True)

            self._merge_dicts(enrolled_data, filial_enrolled)

        self._update_cache(enrolled_data, self.ENROLLED)

        return enrolled_data

    def _collect_capacities(self, dou):

        def _inner(_dou, report_start_date, calc_filial=True):
            capacities = {
                _dou.id: OrderedDict([
                    (INDEX_31, {}),
                    (INDEX_32, {}),
                    (INDEX_32_1, {}),
                ])
            }
            dou_capacities = capacities[_dou.id]

            if not self.has_selected_index(dou_capacities.keys()):
                return capacities

            if calc_filial:
                group_index = Index(dou, report_start_date)
                if not group_index.is_calculated:
                    common = []
                else:
                    common = group_index.free_space_in_unit(
                        group_filter=None,
                        prep_functions=(
                            prepare_enrolled_index_queries,
                            prep_directs_in_group_queries
                        )
                    )

            if INDEX_31 in self.indexes:
                for age_cat in list(const.AGE_CATEGORIES_EIGHT.values()):
                    if calc_filial:
                        dou_capacities[INDEX_31][age_cat] = (
                            childrens_in_age_category(common, age_cat))
                    else:
                        dou_capacities[INDEX_31][age_cat] = []

            direct_index = DirectIndex(dou)

            if INDEX_32 in self.indexes:
                index_32_data = prepare_direct_index_queries(
                    *direct_index.get_queries(*self.list_last_id_directs_32))
                if index_32_data:
                    dou_capacities[INDEX_32][const.ALL] = index_32_data

            if INDEX_32_1 in self.indexes:
                for age_cat in list(const.AGE_CATEGORIES_FULL.values()):
                    index_32_1_data = prepare_direct_index_queries(
                        *direct_index.get_queries(
                            *self.list_last_id_directs_321,
                            age_range=age_cat,
                            on_date=datetime.date.today
                        ))
                    if index_32_1_data:
                        dou_capacities[INDEX_32_1][age_cat] = index_32_1_data

            return capacities

        capacities = _inner(dou, self._report_start_date)
        for filial in self._unit_helper.get_filial(dou).iterator():
            filial_capacities = _inner(
                filial, self._report_start_date, calc_filial=False)

            self._merge_dicts(capacities, filial_capacities)

        self._update_cache(capacities, self.CAPACITIES)

        return capacities

    def _calculate_xml_tags(self, report_form=None):
        """Расчёт тегов в выходном файле xml."""

        unit = self._mo
        region_data = unit.kind.id == UnitKind.REGION

        data_dict = defaultdict(dict)
        for mo in unit.get_descendants(
                include_self=True).filter(kind__id=UnitKind.MO):

            document = TemplateView(
                mo, datetime.datetime.now(),
                ScheduleSettings.get_settings(),
                self._unit_helper,
                report_form,
                region_data=region_data
            )
            # Данные для региона передаются только в одном МО
            region_data = False

            mo_data = self._get_unit_data_from_template_view(document)
            for tag, tag_data_dict in mo_data.items():
                data_dict[tag][mo.id] = tag_data_dict

        return data_dict

    def _get_unit_data_from_template_view(self, document):
        """Получение данных организации из объекта TemplateView."""

        result_dict = defaultdict(dict)
        for dou_data in document.units:

            unit_id = dou_data.unit_id
            unit = get_instance(unit_id, Unit)
            ext_unit = getattr(unit, 'ext_unit', None)

            # Словари с данными
            organization = dou_data.organization
            data_advisory_centr = organization['data_advisory_centr']
            early_assistant_data = organization['early_assistant_data']
            # Маппинг для сопоставления словарей в маппинге XML_TAGS_MAPPING
            data_mapping = {
                'organization': organization,
                'data_advisory_centr': data_advisory_centr,
                'early_assistant_data': early_assistant_data,
            }

            for data_type, tag_info in XML_TAGS_MAPPING.items():
                data = data_mapping[data_type]
                for tag, param_name in tag_info:
                    if tag in self.indexes and param_name in data:
                        result_dict[tag][unit_id] = data[param_name]

            if INDEX_PASSPORT in self.indexes:
                result_dict[INDEX_PASSPORT][unit_id] = (
                    get_detailed_passport(ext_unit))

        return result_dict
