from collections import defaultdict
import os
import uuid
from typing import DefaultDict
from typing import List

from gisdo.collector import ReportCacheMixin
from simple_report.converter.abstract import FileConverter
from simple_report.report import SpreadsheetReport
from simple_report.xlsx.document import DocumentXLSX

from gisdo import settings as gs_settings
from kinder.core.unit.models import Unit
from .collector import ReportDataCollector

from .constants import CAPACITY_TAGS
from .constants import DataTypes
from .constants import INDEXES
from .helpers import ReportRecord


class IndexReportAdapter:
    """Адаптер отчёта "Выгрузка детей по показателю/тегу"."""
    def __init__(self, data: list, indexes: list):
        self.data = data
        self.indexes = indexes

    @staticmethod
    def get_unit_name(unit_id: int) -> str:
        """Получение названия организации по id."""
        return Unit.objects.values_list('name', flat=True).get(id=unit_id)

    @staticmethod
    def get_child_record(
        unit_name, age_cat, child_id, date_of_birth,
        fullname, data_type, data_id
    ):
        """Получение namedtuple строки отчёта для данных ребёнка."""
        data_type_name = DataTypes.values.get(data_type, '')
        return ReportRecord(
            unit_name, age_cat, data_type_name, fullname, data_id)

    @staticmethod
    def get_group_record(group_data, index):
        """Получение namedtuple строки отчёта для данных группы."""
        return ReportRecord(
            group_data['doo_name'],
            '-',
            f'Группа ["{group_data["group_name"]}"]',
            group_data['id'],
            group_data[index],
        )

    @staticmethod
    def get_xml_tag_record(unit_name, value):
        """Получение namedtuple строки отчёта для данных группы."""
        return ReportRecord(unit_name, '', '', '', value)

    def get_data(self) -> DefaultDict[str, List[ReportRecord]]:
        """Получение преобразованных данных."""
        final_data = defaultdict(list)
        groups_data = list()

        for mo_data in self.data:
            for data_type in ReportCacheMixin.DATA_TYPE:
                for unit_id, index_data in mo_data[data_type].items():
                    unit_name = self.get_unit_name(unit_id)
                    if isinstance(index_data, dict):
                        for index, data in index_data.items():
                            child_records = [
                                self.get_child_record(unit_name, age_cat, *row)
                                for age_cat, rows in data.items()
                                for row in rows
                            ]
                            final_data[index].extend(child_records)
                    else:
                        groups_data.extend(index_data)

            # Отдельно обрабатываются теги из выгружаемой xml
            mo_xml_tags_data = mo_data[ReportDataCollector.XML_TAGS]
            for tag_name, tag_data in mo_xml_tags_data.items():
                for mo_id, unit_tag_value in tag_data.items():
                    for unit_id, value in unit_tag_value.items():
                        record = self.get_xml_tag_record(
                            self.get_unit_name(unit_id), value)
                        final_data[tag_name].append(record)

        for index in set(self.indexes) & set(CAPACITY_TAGS):
            final_data[index] = list()
            for group_data in groups_data:
                group_record = self.get_group_record(group_data, index)
                final_data[index].append(group_record)

        return final_data


class IndexReport:
    """Построитель отчета "Выгрузка детей по показателю/тегу"."""

    def __init__(self, data, indexes):
        self.data = data
        self.indexes = indexes
        self.adapter = IndexReportAdapter(data, indexes)

        self.report_name = str(uuid.uuid4())[:16]

        self.template_name = os.path.join(
            os.path.dirname(__file__), "index_report.xlsx")

        self.result_name = os.path.join(
            gs_settings.DOWNLOADS_DIR, self.report_name + ".xlsx")

        self.report = SpreadsheetReport(
            self.template_name, wrapper=DocumentXLSX, type=FileConverter.XLSX)

    def build(self):
        """Построение отчета"""

        header_section = self.report.get_section('header')
        header_section.flush({})

        tag_section = self.report.get_section('tag')
        data_section = self.report.get_section('data')

        # Данные организации
        data = self.adapter.get_data()

        for index, index_name in INDEXES:
            if index in self.indexes:
                tag_section.flush({'tag_index': index_name})

            report_records = data.get(index)
            if not report_records:
                continue

            for report_record in report_records:
                row_data = report_record._asdict()
                data_section.flush(row_data)

        self.report.build(self.result_name)

        return self.report_name
